#!/negi/bin/ruby
#-*- coding: utf-8 -*-

require 'open3'
require 'fileutils'
require 'yaml'
require 'syslog'
require 'socket'

class GPSD
  class LowAccuracyError < StandardError; end
  class IllegalDataError < StandardError; end
  class UnsupportedTypeError < StandardError; end

  GPRMC = Struct.new(:datetime, :time, :status,
                     :lat_raw, :latitude, :latns,
                     :long_raw, :longitude, :longew,
                     :speed, :directon, :date)
  GPGGA = Struct.new(:time,
                     :lat_raw, :latitude, :latns,
                     :long_raw, :longitude, :longew,
                     :quality, :satcount, :accuracy,
                     :aheight, :gheight)
  
  GPSD_WORKDIR   = '/var/lib/gpsd'
  DATETIME_FILE = '/var/lib/gpsd/datetime.log'
  DTSTATUS_FILE = '/var/lib/gpsd/dtstatus.log'
  
  def initialize(device='/dev/ttyACM0')
    Syslog.open('rgpsd')
    @device = device
    @data   = {}
    FileUtils.mkpath(GPSD_WORKDIR)
  end

  def watch
    Open3.popen3("busybox microcom #{@device}"){|sin, sout, serr, wait|
      while line = sout.gets
        begin
          process(parse(line.chomp))
        rescue LowAccuracyError, IllegaldataError, UnsupportedTypeError => e
          Syslog.warning(e.inspect)
        end
      end

      while line = serr.gets
        Syslog.warning("", line.chomp)
      end
    }
  end

  private
  def parse(line)
    type, *ary = line.split(',')
    case type
    when '$GPRMC'
      return parse_gprmc(ary)
    when '$GPGGA'
      return parse_gpgga(ary)
    when '$GPTXT', '$GPGLL', '$GPGSA', '$GPGSV', '$GPVTG'
      printf("> %s: %s\n", type, ary.join(' ')) if $DEBUG
      return nil
    else
      raise UnsupportedTypeError, line
    end
  end

  def process(data)
    return nil if data.nil?
    update_current_data(data)
    case data.class.name
    when 'GPSD::GPRMC'
      dump_file(data)
      set_datetime(data.datetime)
    else
      dump_file(data)
    end
  end

  def update_current_data(data)
    @data[data.class.name.sub(/.*::/, '').to_sym] = data
  end
  
  def dump_file(data)
    File.open(File.join(GPSD_WORKDIR, "#{data.class.name.sub(/.*::/, '')}.yml"), 'w'){|f|
      f.puts(YAML.dump(data))
    }
  end
  
  def parse_gprmc(ary)
    if ary[1] != 'A'
      raise LowAccuracyError, "Status: #{ary[1]}"
    end
    
    rmc = GPRMC.new
    rmc.datetime  = build_datetime(ary[8], ary[0])
    rmc.time      = ary[0]
    rmc.status    = ary[1]
    rmc.lat_raw   = ary[2]
    rmc.latitude  = build_latlong(ary[2])
    rmc.latns     = ary[3]
    rmc.long_raw  = ary[4]
    rmc.longitude = build_latlong(ary[4])
    rmc.longew    = ary[5]
    rmc.speed     = ary[6]
    rmc.directon  = ary[7]
    rmc.date      = ary[8]
    return rmc
  end

  def parse_gpgga(ary)
    if ary[5].to_i < 1
      raise LowAccuracyError, "Quality: #{ary[5]}"
    end
        
    gga = GPGGA.new
    gga.time      = ary[0]
    gga.lat_raw   = ary[1]
    gga.latitude  = build_latlong(ary[1])
    gga.latns     = ary[2]
    gga.long_raw  = ary[3]
    gga.longitude = build_latlong(ary[3])
    gga.longew    = ary[4]
    gga.quality   = ary[5]
    gga.satcount  = ary[6]
    gga.accuracy  = ary[7]
    gga.aheight   = ary[8]
    gga.gheight   = ary[10]
    return gga
  end

  def build_latlong(latlong)
    if latlong.match(/(\d{1,3})(\d{2}\.\d+)/)
      return $1.to_i + $2.to_f / 60
    else
      raise IllegalDataError, "latlong: #{latlong}"
    end
  end
  
  def build_datetime(date, time)
    if date.match(/(\d{2})(\d{2})(\d{2})/)
      date, month, year = $1, $2, $3
    else
      raise IllegalDataError, "date: #{date}"
    end
    if time.match(/(\d{2})(\d{2})(\d{2})\.(\d{2})/)
      hour, min, sec, msec = $1, $2, $3, $4
    else
      raise IllegalDataError, "time: #{time}"
    end
    utc = Time.utc(year.to_i + 2000, month.to_i, date.to_i,
                   hour.to_i, min.to_i, sec.to_i, msec.to_i)
    write_datetime(utc)
    return utc
  end

  def write_datetime(datetime)
    open(DTSTATUS_FILE, 'w'){|f| f.puts(datetime)}
    open(DATETIME_FILE, 'w'){|f| f.puts(datetime)}
  end

  def set_datetime(dt)
    return false if dt.nil?
    cmd = sprintf("date -u -s '%04d-%02d-%02d %02d:%02d:%02d' > /dev/null",
                  dt.year, dt.month, dt.day,
                  dt.hour, dt.min, dt.sec)
    # puts(cmd)
    system(cmd)
  end
end

class GPSServer < GPSD
  GPSD_SOCKET   = File.join(GPSD_WORKDIR, 'gpsd.sock')

  def start
    Thread.start(){ loop{ watch; sleep 5 }}
    File.unlink(GPSD_SOCKET) if File.exist?(GPSD_SOCKET)
    us = UNIXServer.new(GPSD_SOCKET)
    loop {
      Thread.start(us.accept){|sock|
        sock.write(YAML.dump(@data))
        sock.close
      }
    }
  end
end

GPSServer.new.start
