#!/negi/bin/ruby

require 'socket'
require 'yaml'
require 'fileutils'
require 'syslog'

class GPSD
  GPRMC = Struct.new(:datetime, :time, :status,
                     :lat_raw, :latitude, :latns,
                     :long_raw, :longitude, :longew,
                     :speed, :directon, :date)
  GPGGA = Struct.new(:time,
                     :lat_raw, :latitude, :latns,
                     :long_raw, :longitude, :longew,
                     :quality, :satcount, :accuracy,
                     :aheight, :gheight)
end

class WebCAMD
  GPSD_SOCKET      = '/var/lib/gpsd/gpsd.sock'
  LOCATIOND_SOCKET = '/var/lib/locationd/locationd.sock'
  WEBCAMD_DIR      = '/var/lib/webcamd'
  CONF_FILE        = '/conf/webcam.conf'
  WEBCAM_DEFAULT_CONF  = {
    'devices'      => ['/dev/video0'],
    'interval'     => 30,
    'rotate'       => 30,
    '--resolution' => '640x480',
    '--skip'       => '10',
    '--png'        => '9',
    '--log'        => 'syslog'
  }

  def initialize
    Syslog.open('webcamd')
    FileUtils.mkpath(WEBCAMD_DIR)
  end

  def start
    conf = load_conf
    loop {
      time = Time.now.strftime("%Y%m%d_%H%M%S")
      FileUtils.mkpath(File.join(WEBCAMD_DIR, time))
      cleanup_old_contents(conf['rotate'])
      info = get_location
      conf['devices'].each{|dev|
        get_picture(time, info, dev, conf) }
      open(File.join(WEBCAMD_DIR, time, 'info.yml'), 'w'){|yml|
        yml.puts(YAML.dump(info))
      }
      open(File.join(WEBCAMD_DIR, 'latest.txt'), 'w'){|f|
        f.puts(time)
      }
      FileUtils.symlink(File.join(WEBCAMD_DIR, time),
                        File.join(WEBCAMD_DIR, 'latest'),
                        force: true)
      if File.directory?('/mnt/storage')
        begin
          FileUtils.cp_r(File.join(WEBCAMD_DIR, time), '/mnt/storage')
        rescue => e
          Syslog.warning(e.inspect)
        end
      end
      sleep conf['interval']
    }
  end
  
  private
  def load_conf
    open(CONF_FILE){|f|
      return WEBCAM_DEFAULT_CONF.merge(YAML.load(f.read))
    }
  end
  
  def get_location
    info_h = {}
    gsock = UNIXSocket.new(GPSD_SOCKET)
    gconf = YAML.load(gsock.read)
    gsock.close
    asock = UNIXSocket.new(LOCATIOND_SOCKET)
    asock.printf("%f,%f\n", gconf[:GPRMC].latitude.to_f,
                 gconf[:GPRMC].longitude.to_f)
    aconf = YAML.load(asock.read)
    info_h['info'] = sprintf("%s(%.4f) %s:%.4f %s:%.4f %.3fkm/h %.1fm %.1fm",
                             aconf[:address],
                             aconf[:abs],
                             gconf[:GPRMC].latns,
                             gconf[:GPRMC].latitude.to_f,
                             gconf[:GPRMC].longew,
                             gconf[:GPRMC].longitude.to_f,
                             gconf[:GPRMC].speed.to_f,
                             gconf[:GPGGA].aheight.to_f,
                             gconf[:GPGGA].gheight.to_f)
    info_h.merge!(gconf[:GPRMC].to_h)
    info_h.merge!(gconf[:GPGGA].to_h)
    return info_h
  rescue => e
    Syslog.warning(e.inspect)
    return nil
  end

  def cleanup_old_contents(rotate)
    dirs = Dir.glob(File.join(WEBCAMD_DIR, '*_*'))
    if dirs.length > rotate
      FileUtils.remove_entry_secure(dirs.sort[0], true)
    end
  end
  
  def get_picture(time, info, device, conf)
    opts  = conf.collect{|k, v|
      k.match(/^-/) ? [k,v] : nil}.flatten.compact
    devname = File.basename(device)
    system("fswebcam",
           "--log",    "syslog",
           "--device", device,
           "--info",   (info.nil? ? "GPS data not found." : info['info']),
           "--png",    "9",
           "--save",   File.join(WEBCAMD_DIR, time, "#{devname}.png"),
           *opts)
  end
end

WebCAMD.new.start
